# ducktools.env
# MIT License
# 
# Copyright (c) 2024 David C Ellis
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This is used as a bootstrapping script as well as internally
# For this reason it can't use any non-stdlib modules.

from __future__ import annotations

import sys
import os
import os.path


class UnsupportedPlatformError(Exception):
    pass


PACKAGE_SUBFOLDER = "env"

# Folders used internally
CACHEDENV_FOLDERNAME = "caches"
APPLICATION_FOLDERNAME = "application"

MANAGER_FOLDERNAME = "lib"

# Filenames for configuration and catalogue
CONFIG_FILENAME = "config.json"
CATALOGUE_FILENAME = "catalogue.json"


# Store in LOCALAPPDATA for windows, User folder for other operating systems
if sys.platform == "win32":
    # os.path.expandvars will actually import a whole bunch of other modules
    # Try just using the environment.
    if _local_app_folder := os.environ.get("LOCALAPPDATA"):
        if not os.path.isdir(_local_app_folder):
            raise FileNotFoundError(
                f"Could not find local app data folder {_local_app_folder}"
            )
    else:
        raise EnvironmentError(
            "Environment variable %LOCALAPPDATA% "
            "for local application data folder location "
            "not found"
        )
    USER_FOLDER = _local_app_folder
else:
    USER_FOLDER = os.path.expanduser("~")


def get_platform_python(venv_folder):
    if sys.platform == "win32":
        return os.path.join(venv_folder, "Scripts", "python.exe")
    else:
        return os.path.join(venv_folder, "bin", "python")


def get_platform_folder(name):
    if sys.platform == "win32":
        return os.path.join(USER_FOLDER, name)
    else:
        return os.path.join(USER_FOLDER, f".{name}")


class ManagedPaths:
    project_name: str
    project_folder: str
    config_path: str

    manager_folder: str
    pip_zipapp: str
    uv_executable: str
    env_folder: str

    application_folder: str  # Not yet used
    cache_folder: str

    cache_db: str

    build_base: str

    def __init__(self, project_name="ducktools"):
        self.project_name = project_name

        folder_base = os.path.join(self.project_name, PACKAGE_SUBFOLDER)

        self.project_folder = get_platform_folder(folder_base)
        self.config_path = os.path.join(self.project_folder, CONFIG_FILENAME)

        self.manager_folder = os.path.join(self.project_folder, MANAGER_FOLDERNAME)
        self.pip_zipapp = os.path.join(self.manager_folder, "pip.pyz")
        self.uv_executable = os.path.join(
            self.manager_folder,
            "uv.exe" if sys.platform == "win32" else "uv",
        )
        self.env_folder = os.path.join(self.manager_folder, "ducktools-env")

        self.application_folder = os.path.join(self.project_folder, APPLICATION_FOLDERNAME)
        self.cache_folder = os.path.join(self.project_folder, CACHEDENV_FOLDERNAME)
        self.cache_db = os.path.join(self.cache_folder, CATALOGUE_FILENAME)

        self.build_base = os.path.join(self.project_folder, "build")

    @staticmethod
    def get_app_version(versionfile):
        try:
            with open(versionfile, 'r') as f:
                ver = f.read()
        except FileNotFoundError:
            return None
        return ver

    def get_pip_version(self):
        version_file = f"{self.pip_zipapp}.version"
        return self.get_app_version(version_file)

    def get_env_version(self):
        version_file = f"{self.env_folder}.version"
        return self.get_app_version(version_file)

    def get_uv_version(self):
        version_file = f"{self.uv_executable}.version"
        return self.get_app_version(version_file)

    def build_folder(self):
        """
        Get a temporary folder to use for builds
        """
        # Time to use slow stuff
        os.makedirs(self.build_base, exist_ok=True)

        import tempfile
        return tempfile.TemporaryDirectory(dir=self.build_base)


if __name__ == "__main__":
    paths = ManagedPaths("ducktools")
    print(paths.project_folder)
    print(paths.config_path)
    print(paths.cache_folder)
    print(paths.cache_db)



# ducktools.env
# MIT License
# 
# Copyright (c) 2024 David C Ellis
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This script becomes the __main__.py script inside the ducktools.pyz zipapp.
"""
from __future__ import annotations

import os.path
import sys
import zipfile

# This file is moved and these imports created when bundling
from _platform_paths import ManagedPaths  # type: ignore

from _vendor.ducktools.lazyimporter import LazyImporter, FromImport, ModuleImport  # type: ignore
from _vendor import zipp  # type: ignore

# This bootstrap script exists without ducktools.env and so needs a copy of project_name
PROJECT_NAME = "ducktools"

_laz = LazyImporter(
    [
        FromImport("_vendor.packaging.version", "Version"),
        ModuleImport("runpy"),
        ModuleImport("shutil"),
    ]
)

default_paths = ManagedPaths(PROJECT_NAME)


def is_outdated(installed_version: str | None, bundled_version: str) -> bool:
    # Shortcut for no version installed
    if installed_version is None:
        return True

    # Always consider dev versions outdated
    if "dev" in installed_version:
        return True

    # Shortcut for identical version string
    if installed_version == bundled_version:
        return False

    # Try to use tuples, fallback to packaging
    try:
        installed_info = tuple(int(segment) for segment in installed_version.split("."))
        bundled_info = tuple(int(segment) for segment in bundled_version.split("."))
    except (ValueError, TypeError):
        installed_info = _laz.Version(installed_version)
        bundled_info = _laz.Version(bundled_version)

    return installed_info < bundled_info


def update_libraries():
    archive_path = os.path.abspath(sys.argv[0])

    # Compare library versions to those in cache
    with zipfile.ZipFile(archive_path, "r") as zf:
        bundled_ducktools_ver = zipp.Path(zf, "ducktools-env.version").read_text()
        bundled_pip_ver = zipp.Path(zf, "pip.pyz.version").read_text()

        # Copy ducktools if outdated
        if is_outdated(default_paths.get_env_version(), bundled_ducktools_ver):
            sys.stderr.write("Installed ducktools is older than bundled, replacing.\n")
            extract_names = sorted(n for n in zf.namelist() if n.startswith("ducktools-env/"))
            zf.extractall(default_paths.manager_folder, members=extract_names)
            zf.extract("ducktools-env.version", default_paths.manager_folder)

        # Copy pip if outdated
        if is_outdated(default_paths.get_pip_version(), bundled_pip_ver):
            sys.stderr.write("Installed pip is older than bundled, replacing.\n")
            zf.extract("pip.pyz", default_paths.manager_folder)
            zf.extract("pip.pyz.version", default_paths.manager_folder)


def launch_script(script_file, zipapp_path, args, lockdata=None):
    sys.path.insert(0, default_paths.env_folder)
    try:
        from ducktools.env.manager import Manager
        from ducktools.env.environment_specs import EnvironmentSpec
        manager = Manager(PROJECT_NAME)
        spec = EnvironmentSpec.from_script(
            script_path=script_file,
            lockdata=lockdata,
        )

        manager.run_bundled_script(
            spec=spec,
            zipapp_path=zipapp_path,
            args=args,
        )
    finally:
        sys.path.pop(0)


def launch_ducktools():
    _laz.runpy.run_path(default_paths.env_folder, run_name="__main__")



# ducktools.env
# MIT License
# 
# Copyright (c) 2024 David C Ellis
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import os.path
import shutil
import subprocess
import sys

from ducktools.env.platform_paths import ManagedPaths
from ducktools.env.scripts.get_pip import retrieve_pip

uv_versionspec = "~=0.4.0"

uv_download = "bin/uv.exe" if sys.platform == "win32" else "bin/uv"


def retrieve_uv(paths: ManagedPaths) -> str | None:
    if os.path.exists(paths.uv_executable):
        uv_path = paths.uv_executable
    else:
        pip_install = retrieve_pip(paths=paths)
        with paths.build_folder() as build_folder:

            install_folder = os.path.join(build_folder, "uv")

            uv_dl = os.path.join(install_folder, uv_download)

            pip_command = [
                sys.executable,
                pip_install,
                "--disable-pip-version-check",
                "install",
                "-q",
                f"uv{uv_versionspec}",
                "--only-binary=:all:",
                "--target",
                install_folder,
            ]

            # Download UV with pip - handles getting the correct platform version
            try:
                subprocess.run(
                    pip_command,
                    check=True,
                )
            except subprocess.CalledProcessError:
                uv_path = None
            else:
                # Copy the executable out of the pip install
                shutil.copy(uv_dl, paths.uv_executable)
                uv_path = paths.uv_executable

                version_command = [uv_path, "-V"]
                version_output = subprocess.run(version_command, capture_output=True, text=True)
                uv_version = version_output.stdout.split()[1]
                with open(f"{uv_path}.version", 'w') as ver_file:
                    ver_file.write(uv_version)

    return uv_path



# DuckTools-EnvMan
# Copyright (C) 2024 David C Ellis
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import unittest.mock as mock

from ducktools.env.environment_specs import EnvironmentSpec
from ducktools.env.environment_specs import _laz  # noqa

from ducktools.classbuilder.prefab import prefab, attribute
import pytest

from packaging.specifiers import SpecifierSet
from packaging.requirements import Requirement


MOCK_RUN_STDOUT = "<MOCK DATA>"


@pytest.fixture
def subprocess_run_mock():
    with mock.patch.object(_laz, "subprocess") as sp:
        run_mock = mock.MagicMock()
        run_return_mock = mock.MagicMock()
        run_return_mock.stdout = MOCK_RUN_STDOUT
        run_mock.return_value = run_return_mock

        sp.run = run_mock
        yield run_mock


@prefab
class DataSet:
    raw_spec: str
    requires_python: str | None = None
    dependencies: list[str] = attribute(default_factory=list)
    extras: dict = attribute(default_factory=dict)


envs = [
    DataSet(
        raw_spec="",
    ),
    DataSet(
        raw_spec=("requires-python = '>=3.10'\n" "dependencies = []\n"),
        requires_python=">=3.10",
    ),
    DataSet(
        raw_spec=(
            "requires-python = '>=3.11'\n" "dependencies = ['ducktools-env>=0.1.0']\n"
        ),
        requires_python=">=3.11",
        dependencies=["ducktools-env>=0.1.0"],
    ),
]


@pytest.mark.parametrize("test_data", envs)
def test_envspec_pythononly(test_data):
    env = EnvironmentSpec(
        "path/to/script.py",
        test_data.raw_spec
    )

    assert env.details.requires_python == test_data.requires_python
    assert env.details.dependencies == test_data.dependencies


@pytest.mark.parametrize("test_data", envs)
def test_generate_lockdata(test_data, subprocess_run_mock):
    env = EnvironmentSpec(
        "path/to/script.py",
        test_data.raw_spec,
    )
    fake_uv_path = "fake/uv/path"

    lock_data = env.generate_lockdata(fake_uv_path)

    if test_data.dependencies:
        deps = "\n".join(env.details.dependencies)
        # Check the mock output is there
        assert lock_data == MOCK_RUN_STDOUT

        # Check the mock is called correctly
        subprocess_run_mock.assert_called_once_with(
            [
                fake_uv_path,
                "pip",
                "compile",
                "--universal",
                "--generate-hashes",
                "--no-annotate",
                "--python-version",
                "3.11",
                "-",
            ],
            input=deps,
            capture_output=True,
            text=True,
        )

    else:
        # No dependencies, shouldn't call subprocess
        subprocess_run_mock.assert_not_called()

        assert lock_data == "# No Dependencies Declared"


@pytest.mark.parametrize("test_data", envs)
def test_requires_python_spec(test_data):
    # Test that the requires_python_spec function returns the correct specifierset
    env = EnvironmentSpec(
        "path/to/script.py",
        test_data.raw_spec,
    )

    if test_data.requires_python:
        assert env.details.requires_python_spec == SpecifierSet(
            test_data.requires_python
        )
    else:
        assert env.details.requires_python_spec is None


@pytest.mark.parametrize("test_data", envs)
def test_dependencies_spec(test_data):
    env = EnvironmentSpec(
        "path/to/script.py",
        test_data.raw_spec,
    )

    assert env.details.dependencies_spec == [
        Requirement(s) for s in test_data.dependencies
    ]


def test_spec_errors():
    fake_spec = (
        "requires-python = '!!>=3.10'\n"
        "dependencies = ['invalid_spec!', 'valid_spec>=3.10']\n"
    )

    env = EnvironmentSpec(
        "path/to/script.py",
        fake_spec,
    )

    errs = env.details.errors()

    assert errs == [
        "Invalid python version specifier: '!!>=3.10'",
        "Invalid dependency specification: 'invalid_spec!'",
    ]


@pytest.mark.parametrize("test_data", envs)
def test_asdict(test_data):
    env = EnvironmentSpec(
        "path/to/script.py",
        test_data.raw_spec,
    )

    assert env.as_dict() == {
        "spec_hash": env.spec_hash,
        "raw_spec": test_data.raw_spec,
        "details": {
            "requires_python": test_data.requires_python,
            "dependencies": test_data.dependencies,
            "tool_table": {},
        },
        "lock_hash": None,
    }



# DuckTools-EnvMan
# Copyright (C) 2024 David C Ellis
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os.path
import shutil

from ducktools.pythonfinder import get_python_installs
from ducktools.pythonfinder.shared import get_install_details

from ducktools.env.catalogue import TempCatalogue
from ducktools.env.config import Config

import ducktools.env.platform_paths as platform_paths

from unittest.mock import patch
import pytest


@pytest.fixture(scope="session")
def available_pythons():
    return get_python_installs()


@pytest.fixture(scope="session")
def this_python():
    py = sys.executable
    return get_install_details(py)


@pytest.fixture(scope="session", autouse=True)
def use_this_python_install(this_python):
    with patch("ducktools.env.catalogue._laz.list_python_installs") as get_installs:
        get_installs.return_value = [this_python]
        yield


@pytest.fixture(scope="function")
def catalogue_path():
    """
    Provide a test folder path for python environments, delete after tests in a class have run.
    """
    folder = os.path.join(os.path.dirname(__file__), "test_envs")
    cache_file = os.path.join(folder, platform_paths.CATALOGUE_FILENAME)
    yield cache_file
    try:
        shutil.rmtree(folder)
    except FileNotFoundError:
        pass


@pytest.fixture(scope="session")
def test_config():
    config = Config(
        cache_maxcount=2,
        cache_lifetime=1/24,
    )
    yield config


@pytest.fixture(scope="function")
def testing_catalogue(catalogue_path):
    catalogue = TempCatalogue(path=catalogue_path)
    yield catalogue



# ducktools.env
# MIT License
# 
# Copyright (c) 2024 David C Ellis
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module handles downloading and installing the `pip.pyz` zipapp if it is outdated
or not installed.

The pip zipapp will be included in ducktools.pyz builds so this should only be needed
when building ducktools.pyz or if ducktools-env has been installed via pip.
"""

from __future__ import annotations

import os
import os.path

from ducktools.classbuilder.prefab import prefab
from ducktools.lazyimporter import LazyImporter, FromImport, ModuleImport

from ducktools.env.platform_paths import ManagedPaths
from ducktools.env.config import log
from ducktools.env.exceptions import InvalidPipDownload


BASE_URL = "https://bootstrap.pypa.io/pip"


_laz = LazyImporter(
    [
        ModuleImport("hashlib"),
        FromImport("urllib.request", "urlopen"),
        FromImport("packaging.version", "Version"),
    ]
)


@prefab(frozen=True)
class PipZipapp:
    version_str: str
    sha3_256: str
    source_url: str

    @property
    def full_url(self):
        return f"{BASE_URL}/{self.source_url}"

    @property
    def version_tuple(self):
        return tuple(int(segment) for segment in self.version_str.split("."))

    @property
    def as_version(self):
        return _laz.Version(self.version_str)


# This is mostly kept for testing.
PREVIOUS_PIP = PipZipapp(
    version_str="24.1",
    sha3_256="f4c4d76e70498762832a842d5b55e9d8c09b6d6607b30b5f4eb08e68dfc57077",
    source_url="zipapp/pip-24.1.pyz"
)

LATEST_PIP = PipZipapp(
    version_str="24.2",
    sha3_256="8dc4860613c47cb2e5e55c7e1ecf4046abe18edca083073d51f1720011bed6ea",
    source_url="zipapp/pip-24.2.pyz",
)


def is_pip_outdated(
        paths: ManagedPaths,
        latest_version: PipZipapp = LATEST_PIP
):
    pip_version = paths.get_pip_version()

    if pip_version is None:
        return True

    try:
        installed_info = tuple(int(segment) for segment in pip_version.split("."))
        latest_info = latest_version.version_tuple
    except (ValueError, TypeError):
        # possible pre/post release versions - use packaging
        installed_info = _laz.Version(pip_version)
        latest_info = latest_version.as_version

    return installed_info < latest_info


def download_pip(
        pip_destination: str,
        latest_version: PipZipapp = LATEST_PIP
):

    url = latest_version.full_url

    # Actual download
    with _laz.urlopen(url) as f:
        data = f.read()

    # Check hash matches
    if _laz.hashlib.sha3_256(data).hexdigest() != latest_version.sha3_256:
        raise InvalidPipDownload(
            "The checksum of the downloaded PIP binary did not match the expected value."
        )

    # Make directory if it does not exist
    os.makedirs(os.path.dirname(pip_destination), exist_ok=True)

    with open(pip_destination, 'wb') as f:
        f.write(data)

    with open(f"{pip_destination}.version", 'w') as f:
        f.write(".".join(str(item) for item in latest_version.version_tuple))


def retrieve_pip(
    paths: ManagedPaths,
    latest_version: PipZipapp = LATEST_PIP,
) -> str:
    """
    If pip.pyz is not installed, download it and place it in the cache
    return the path to the .pyz

    :param paths:
    :param latest_version:
    :return: path to pip.pyz
    """

    if is_pip_outdated(paths, latest_version=latest_version):
        log("Downloading PIP")
        download_pip(paths.pip_zipapp, latest_version=latest_version)

        log(f"Pip zipapp installed at {paths.pip_zipapp!r}")

    return paths.pip_zipapp



# /// script
# requires-python = ">=3.11"
# dependencies = [
#   "requests<3",
#   "rich",
# ]
# ///

import requests
from rich.pretty import pprint

resp = requests.get("https://peps.python.org/api/peps.json")
data = resp.json()
pprint([(k, v["title"]) for k, v in data.items()][:10])



# ducktools.env
# MIT License
# 
# Copyright (c) 2024 David C Ellis
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

"""
This is the script that builds the inner ducktools-env folder
and bundles ducktools-env into ducktools.pyz
"""
import sys
import os
import os.path
import shutil
import subprocess
import zipapp

from pathlib import Path

import importlib_resources
from importlib.metadata import requires
from packaging.requirements import Requirement

import ducktools.env
from ducktools.env import MINIMUM_PYTHON_STR, bootstrap_requires
from ducktools.env.platform_paths import ManagedPaths


def build_env_folder(
        *,
        paths: ManagedPaths,
        install_base_command: list[str],
        clear_old_builds=True
) -> None:
    # Get the full requirements for ducktools-env
    deps = []
    reqs = requires("ducktools-env")  # Use hyphen name to be recognised by older python
    for req in reqs:
        req = Requirement(req)
        if not (req.marker and not req.marker.evaluate({"python_version": MINIMUM_PYTHON_STR})):
            deps.append(f"{req.name}{req.specifier}")

    with paths.build_folder() as build_folder:

        if clear_old_builds:
            build_folder_path = Path(build_folder)
            for p in build_folder_path.parent.glob("*"):
                if p != build_folder_path:
                    shutil.rmtree(p)

        print("Downloading application dependencies")

        # install packages into build folder
        install_command = [
            *install_base_command,
            "install",
            *deps,
            "--python-version",
            MINIMUM_PYTHON_STR,
            "--only-binary=:all:",
            "--no-compile",
            "--target",
            build_folder,
        ]

        subprocess.run(install_command)

        freeze_command = [
            *install_base_command,
            "freeze",
            "--path",
            build_folder,
        ]

        # don't include executable scripts
        shutil.rmtree(os.path.join(build_folder, "bin"), ignore_errors=True)

        freeze = subprocess.run(freeze_command, capture_output=True, text=True)

        (Path(build_folder) / "requirements.txt").write_text(freeze.stdout)

        # Get the paths for modules that need to be copied
        resources = importlib_resources.files("ducktools.env")

        with importlib_resources.as_file(resources) as env_folder:
            print("Copying application into archive")
            ignore_compiled = shutil.ignore_patterns("__pycache__")
            shutil.copytree(
                env_folder,
                os.path.join(build_folder, "ducktools", "env"),
                ignore=ignore_compiled,
            )

            main_app_path = env_folder / "__main__.py"
            print("Copying __main__.py into lib")
            shutil.copy(main_app_path, build_folder)

        print("Creating ducktools-env lib folder")
        shutil.rmtree(paths.env_folder, ignore_errors=True)
        shutil.copytree(
            build_folder,
            paths.env_folder,
        )

    print("Writing env version number")
    with open(paths.env_folder + ".version", 'w') as f:
        f.write(ducktools.env.__version__)


def build_zipapp(
        *,
        paths: ManagedPaths,
        install_base_command: list[str],
        clear_old_builds=True
) -> None:
    archive_name = f"ducktools.pyz"

    with paths.build_folder() as build_folder:

        if clear_old_builds:
            build_folder_path = Path(build_folder)
            for p in build_folder_path.parent.glob("*"):
                if p != build_folder_path:
                    shutil.rmtree(p)

        # UV should not be bundled - binary is not cross platform
        uv_base_exe = "uv.exe" if sys.platform == "win32" else "uv"
        uv_pattern = shutil.ignore_patterns(uv_base_exe, f"{uv_base_exe}.version")

        print("Copying pip.pyz and ducktools-env")
        shutil.copytree(paths.manager_folder, build_folder, ignore=uv_pattern, dirs_exist_ok=True)

        # Get the paths for modules that need to be copied
        resources = importlib_resources.files("ducktools.env")

        with importlib_resources.as_file(resources) as env_folder:
            platform_paths_path = env_folder / "platform_paths.py"
            bootstrap_path = env_folder / "bootstrapping" / "bootstrap.py"
            main_zipapp_path = env_folder / "bootstrapping" / "zipapp_main.py"

            print("Copying platform paths")
            shutil.copy(platform_paths_path, os.path.join(build_folder, "_platform_paths.py"))

            print("Copying bootstrap script")
            shutil.copy(bootstrap_path, os.path.join(build_folder, "_bootstrap.py"))

            print("Copying __main__ script")
            shutil.copy(main_zipapp_path, os.path.join(build_folder, "__main__.py"))

        print("Installing bootstrap requirements")
        vendor_folder = os.path.join(build_folder, "_vendor")

        pip_command = [
            *install_base_command,
            "install",
            *bootstrap_requires,
            "--python-version",
            MINIMUM_PYTHON_STR,
            "--only-binary=:all:",
            "--no-compile",
            "--target",
            vendor_folder,
        ]
        subprocess.run(pip_command)

        freeze_command = [
            *install_base_command,
            "freeze",
            "--path",
            vendor_folder,
        ]

        freeze = subprocess.run(freeze_command, capture_output=True, text=True)

        (Path(vendor_folder) / "requirements.txt").write_text(freeze.stdout)

        dist_folder = Path(os.getcwd(), "dist")
        dist_folder.mkdir(exist_ok=True)

        print(f"Creating {archive_name}")
        zipapp.create_archive(
            source=build_folder,
            target=dist_folder / archive_name,
            interpreter="/usr/bin/env python"
        )



# ducktools.env
# MIT License
# 
# Copyright (c) 2024 David C Ellis
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
User global configuration
"""
import sys
import os

from datetime import timedelta as _timedelta

from ducktools.lazyimporter import LazyImporter, ModuleImport
from ducktools.classbuilder.prefab import Prefab, get_attributes, as_dict

_laz = LazyImporter(
    [
        ModuleImport("json"),
    ]
)


def log(message):
    sys.stderr.write(message)
    sys.stderr.write("\n")


class Config(Prefab, kw_only=True):
    # Global settings for caches
    cache_maxcount: int = 2
    cache_lifetime: float = 1.0

    applications_expire: bool = False
    applications_lifetime: float = 28.0

    # Use uv and allow uv to auto install Python
    use_uv: bool = True
    uv_install_python: bool = True  # Not yet functional

    @property
    def cache_lifetime_delta(self) -> _timedelta:
        return _timedelta(days=self.cache_lifetime)

    @property
    def application_lifetime_delta(self) -> _timedelta:
        return _timedelta(days=self.applications_lifetime)

    @classmethod
    def load(cls, file_path: str):
        try:
            with open(file_path, 'r') as f:
                json_data = _laz.json.load(f)
        except (FileNotFoundError, _laz.json.JSONDecodeError):
            return cls()
        else:
            attribute_keys = {k for k, v in get_attributes(cls).items() if v.init}

            filtered_data = {
                k: v for k, v in json_data.items() if k in attribute_keys
            }

            # noinspection PyArgumentList
            return cls(**filtered_data)

    def save(self, file_path: str):
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w") as f:
            _laz.json.dump(self, f, default=as_dict)



# ducktools.env
# MIT License
#
# Copyright (c) 2024 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import sys
import os.path
from datetime import datetime as _datetime, timedelta as _timedelta

from ducktools.lazyimporter import (
    LazyImporter,
    FromImport,
    ModuleImport,
    MultiFromImport,
)

from ducktools.classbuilder.prefab import Prefab, prefab, attribute, as_dict, get_attributes

from .exceptions import PythonVersionNotFound, InvalidEnvironmentSpec, VenvBuildError
from .environment_specs import EnvironmentSpec
from .config import Config, log


_laz = LazyImporter(
    [
        ModuleImport("shutil"),
        ModuleImport("json"),
        ModuleImport("subprocess"),
        ModuleImport("hashlib"),
        ModuleImport("tempfile"),
        FromImport("ducktools.pythonfinder", "list_python_installs"),
    ],
    globs=globals(),
)

_packaging = LazyImporter(
    [
        MultiFromImport(
            "packaging.requirements",
            ["Requirement", "InvalidRequirement"],
        ),
        MultiFromImport(
            "packaging.specifiers",
            ["SpecifierSet", "InvalidSpecifier"],
        ),
        MultiFromImport("packaging.version", ["Version", "InvalidVersion"]),
    ]
)


def _datetime_now_iso() -> str:
    """
    Helper function to allow use of datetime.now with iso formatting
    as a default factory
    """
    return _datetime.now().isoformat()


class BaseEnv(Prefab, kw_only=True):
    name: str
    path: str
    python_version: str
    parent_python: str
    created_on: str = attribute(default_factory=_datetime_now_iso)
    last_used: str = attribute(default_factory=_datetime_now_iso)

    @property
    def python_path(self) -> str:
        if sys.platform == "win32":
            return os.path.join(self.path, "Scripts", "python.exe")
        else:
            return os.path.join(self.path, "bin", "python")

    @property
    def created_date(self) -> _datetime:
        return _datetime.fromisoformat(self.created_on)

    @property
    def last_used_date(self) -> _datetime:
        return _datetime.fromisoformat(self.last_used)

    @property
    def exists(self) -> bool:
        return os.path.exists(self.python_path)

    @property
    def parent_exists(self) -> bool:
        return os.path.exists(self.parent_python)

    @property
    def is_valid(self) -> bool:
        """Check that both the folder exists and the source python exists"""
        return self.exists and self.parent_exists

    def delete(self) -> None:
        """Delete the cache folder"""
        _laz.shutil.rmtree(self.path)


class TemporaryEnv(BaseEnv, kw_only=True):
    """
    This is for temporary environments that expire after a certain period
    """
    spec_hashes: list[str]
    lock_hash: str | None = None
    installed_modules: list[str] = attribute(default_factory=list)


class ApplicationEnv(BaseEnv, kw_only=True):
    ...


@prefab(kw_only=True)
class BaseCatalogue:
    ENV_TYPE = BaseEnv

    path: str
    environments: dict[str, BaseEnv] = attribute(default_factory=dict)

    @property
    def catalogue_folder(self):
        return os.path.dirname(self.path)

    def save(self) -> None:
        """Serialize this class into a JSON string and save"""
        # For external users that may not import prefab directly
        os.makedirs(self.catalogue_folder, exist_ok=True)

        with open(self.path, "w") as f:
            _laz.json.dump(self, f, default=as_dict, indent=2)

    @classmethod
    def load(cls, path):
        try:
            with open(path, 'r') as f:
                json_data = _laz.json.load(f)
        except (FileNotFoundError, _laz.json.JSONDecodeError):
            # noinspection PyArgumentList
            return cls(path=path)
        else:
            cls_keys = {k for k, v in get_attributes(cls).items() if v.init}

            filtered_data = {
                k: v for k, v in json_data.items() if k in cls_keys
            }

            environments = {}
            for k, v in filtered_data.get("environments", {}).items():
                environments[k] = cls.ENV_TYPE(**v)

            filtered_data["environments"] = environments

            # noinspection PyArgumentList
            return cls(**filtered_data)

    def delete_env(self, envpath: str) -> None:
        if env := self.environments.get(envpath):
            env.delete()
            del self.environments[envpath]
            self.save()
        else:
            raise FileNotFoundError(f"Cache {envpath!r} not found")

    def purge_folder(self):
        """
        Clear the cache folder when things have gone wrong or for a new version.
        """
        # This does not save as the act of deleting the catalogue folder 
        # will delete the file. It should not automatically be recreated.

        # Clear the folder
        try:
            _laz.shutil.rmtree(self.catalogue_folder)
        except FileNotFoundError:  # pragma: no cover
            pass

        # Clear environment list
        self.environments = {}


@prefab(kw_only=True)
class TempCatalogue(BaseCatalogue):
    """
    Catalogue for temporary environments
    """
    ENV_TYPE = TemporaryEnv

    environments: dict[str, TemporaryEnv] = attribute(default_factory=dict)
    env_counter: int = 0

    @property
    def oldest_cache(self) -> str | None:
        """
        :return: name of the oldest cache or None if there are no caches
        """
        old_cache = None
        for cache in self.environments.values():
            if old_cache:
                if cache.last_used < old_cache.last_used:
                    old_cache = cache
            else:
                old_cache = cache

        if old_cache:
            return old_cache.name
        else:
            return None

    def expire_caches(self, lifetime: _timedelta) -> None:
        """
        Delete caches that are older than `lifetime`

        :param lifetime: timedelta age after which caches should be deleted
        :type lifetime: _timedelta
        """
        if lifetime:
            ctime = _datetime.now()
            # Iterate over a copy as we are modifying the original
            for cachename, cache in self.environments.copy().items():
                if (ctime - cache.created_date) > lifetime:
                    self.delete_env(cachename)

        self.save()

    def find_env_hash(self, *, spec: EnvironmentSpec) -> TemporaryEnv | None:
        """
        Attempt to find a cached python environment that matches the hash
        of the specification.

        This means that either the exact text was used to generate the environment
        or that it has previously matched in sufficient mode.

        :param spec: EnvironmentSpec of requirements
        :return: CacheFolder details of python env that satisfies it or None
        """
        for cache in self.environments.values():
            if spec.spec_hash in cache.spec_hashes:
                log(f"Hash {spec.spec_hash!r} matched environment {cache.name}")

                if not cache.is_valid:
                    log(f"Cache {cache.name!r} does not point to a valid python, removing.")
                    self.delete_env(cache.name)
                    continue

                cache.last_used = _datetime_now_iso()
                self.save()
                return cache
        else:
            return None

    def find_sufficient_env(self, *, spec: EnvironmentSpec) -> TemporaryEnv | None:
        """
        Check for a cache that matches the minimums of all specified modules

        If found, add the text of the spec to raw_specs for that module and return it.

        :param spec: EnvironmentSpec requirements for a python environment
        :return: TemporaryEnv environment or None
        """

        for cache in self.environments.values():
            # If no python version listed ignore it
            # If python version is listed, make sure it matches
            if spec.details.requires_python:
                cache_pyver = _packaging.Version(cache.python_version)
                if not spec.details.requires_python_spec.contains(cache_pyver, prereleases=True):
                    continue

            # Check dependencies
            cache_spec = {}

            for mod in cache.installed_modules:
                name, version = mod.split("==")
                # There should only be one specifier, specifying one version
                module_ver = _packaging.Version(version)
                cache_spec[name] = module_ver

            for req in spec.details.dependencies_spec:
                # If a dependency is not satisfied , break out of this loop
                if ver := cache_spec.get(req.name):
                    if ver not in req.specifier:
                        break
                else:
                    break
            else:
                # If all dependencies were satisfied, the loop completed
                # Update last_used and append the hash of the spec to the spec hashes
                log(f"Spec satisfied by {cache.name!r}")

                if not cache.is_valid:
                    log(f"Cache {cache.name!r} does not point to a valid python, removing.")
                    self.delete_env(cache.name)
                    continue

                log(f"Adding {spec.spec_hash!r} to {cache.name!r} hash list")

                cache.last_used = _datetime_now_iso()
                cache.spec_hashes.append(spec.spec_hash)

                self.save()
                return cache

        else:
            return None

    def find_locked_env(
        self, 
        *, 
        spec: EnvironmentSpec,
    ) -> TemporaryEnv | None:
        """
        Find a cached TemporaryEnv that matches the hash of the lockfile

        :param spec: Environment specification (needed for lock)
        :return: TemporaryEnv environment or None
        """
        # Get lock data hash
        for cache in self.environments.values():
            if (
                cache.lock_hash == spec.lock_hash 
                and cache.python_version in spec.details.requires_python_spec
            ):
                log(f"Lockfile hash {spec.lock_hash!r} matched environment {cache.name}")
                return cache
        else:
            return None

    def find_env(self, *, spec: EnvironmentSpec) -> TemporaryEnv | None:
        """
        Try to find an existing cached environment that satisfies the spec

        :param spec: Environment specification
        :return: TemporaryEnv environment or None
        """
        if spec.lock_hash:
            env = self.find_locked_env(spec=spec)
        
        else:
            env = self.find_env_hash(spec=spec)

            if not env:
                env = self.find_sufficient_env(spec=spec)

        return env

    def create_env(
        self,
        *,
        spec: EnvironmentSpec,
        config: Config,
        uv_path: str | None,
        installer_command: list[str],
    ) -> TemporaryEnv:
        # Check the spec is valid
        if spec_errors := spec.details.errors():
            raise InvalidEnvironmentSpec("; ".join(spec_errors))

        # Delete the oldest cache if there are too many
        while len(self.environments) >= config.cache_maxcount:
            del_cache = self.oldest_cache
            log(f"Deleting {del_cache}")
            self.delete_env(del_cache)

        new_cachename = f"env_{self.env_counter}"
        self.env_counter += 1

        cache_path = os.path.join(self.catalogue_folder, new_cachename)

        # Find a valid python executable
        for install in _laz.list_python_installs():
            if install.implementation.lower() != "cpython":
                # Ignore all non cpython installs for now
                continue
            if (
                not spec.details.requires_python
                or spec.details.requires_python_spec.contains(install.version_str, prereleases=True)
            ):
                python_exe = install.executable
                python_version = install.version_str
                break
        else:
            raise PythonVersionNotFound(
                f"Could not find a Python install satisfying {spec.details.requires_python!r}."
            )

        # Make the venv
        if os.path.exists(cache_path):
            raise FileExistsError(
                f"Cache path {cache_path!r} already exists. "
                f"Clear caches to resolve."
            )

        try:
            log(f"Creating venv in: {cache_path}")
            if uv_path:
                _laz.subprocess.run(
                    [uv_path, "venv", "-q", "--python", python_exe, cache_path], check=True
                )
            else:
                _laz.subprocess.run(
                    [python_exe, "-m", "venv", "--without-pip", cache_path], check=True
                )
        except _laz.subprocess.CalledProcessError as e:
            # Try to delete the folder if it exists
            _laz.shutil.rmtree(cache_path, ignore_errors=True)
            raise VenvBuildError(f"Failed to build venv: {e}")

        # Construct the TemporaryEnv to get the path to python.exe
        new_env = TemporaryEnv(
            name=new_cachename,
            path=cache_path,
            python_version=python_version,
            parent_python=python_exe,
            spec_hashes=[spec.spec_hash],
            lock_hash=spec.lock_hash,
        )

        if deps := spec.details.dependencies:
            dep_list = ", ".join(deps)
            log(f"Installing dependencies from PyPI: {dep_list}")

            if spec.lockdata:
                log("Using lockfile")
                # Need a temporary file to use as the lockfile
                with _laz.tempfile.TemporaryDirectory() as tempfld:
                    requirements_path = os.path.join(tempfld, "requirements.txt")
                    with open(requirements_path, 'w') as f:
                        f.write(spec.lockdata)
                    try:
                        if uv_path:
                            dependency_command = [
                                *installer_command,
                                "install",
                                "-q",  # Quiet
                                "--python",
                                new_env.python_path,
                                "-r",
                                requirements_path,
                            ]
                        else:
                            dependency_command = [
                                *installer_command,
                                "--python",
                                new_env.python_path,
                                "install",
                                "-q",  # Quiet
                                "-r", 
                                requirements_path,
                            ]
                        _laz.subprocess.run(
                            dependency_command,
                            check=True,
                        )
                    except _laz.subprocess.CalledProcessError as e:
                        raise VenvBuildError(f"Failed to install dependencies: {e}")
            else:
                try:
                    if uv_path:
                        dependency_command = [
                            *installer_command,
                            "install",
                            "-q",  # Quiet
                            "--python",
                            new_env.python_path,
                            *deps,
                        ]
                    else:
                        dependency_command = [
                            *installer_command,
                            "--python",
                            new_env.python_path,
                            "install",
                            "-q",  # Quiet
                            *deps,
                        ]
                    _laz.subprocess.run(
                        dependency_command,
                        check=True,
                    )
                except _laz.subprocess.CalledProcessError as e:
                    # Try to delete the folder if it exists
                    _laz.shutil.rmtree(cache_path, ignore_errors=True)
                    raise VenvBuildError(f"Failed to install dependencies: {e}")

            # Get pip-freeze list to use for installed modules
            if uv_path:
                freeze_command = [
                    *installer_command,
                    "freeze",
                    "--python",
                    new_env.python_path,
                ]
            else:
                freeze_command = [
                    *installer_command,
                    "--python",
                    new_env.python_path,
                    "freeze",
                ]
            freeze = _laz.subprocess.run(
                freeze_command,
                capture_output=True,
                text=True,
            )

            installed_modules = [
                item.strip()
                for item in freeze.stdout.splitlines()
                if item
            ]

            new_env.installed_modules.extend(installed_modules)

        self.environments[new_cachename] = new_env
        self.save()

        return new_env

    def find_or_create_env(
        self,
        *,
        spec: EnvironmentSpec,
        config: Config,
        uv_path: str | None,
        installer_command: list[str],
    ) -> TemporaryEnv:
        env = self.find_env(spec=spec)

        if not env:
            log("Existing environment not found, creating new environment.")
            env = self.create_env(
                spec=spec,
                config=config,
                uv_path=uv_path,
                installer_command=installer_command,
            )
        return env



# ducktools.env
# MIT License
# 
# Copyright (c) 2024 David C Ellis
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

from ducktools.lazyimporter import (
    LazyImporter,
    FromImport,
    ModuleImport,
    MultiFromImport,
    TryExceptImport,
)

from ducktools.classbuilder.prefab import Prefab, as_dict
import ducktools.scriptmetadata as scriptmetadata


# Lazy importers for modules that may not be used
_laz = LazyImporter(
    [
        TryExceptImport(
            "tomllib",
            "tomli",
            "tomllib",
        ),
        ModuleImport("warnings"),
        ModuleImport("subprocess"),
        ModuleImport("hashlib"),
        MultiFromImport(
            "packaging.requirements",
            ["Requirement", "InvalidRequirement"],
        ),
        MultiFromImport(
            "packaging.specifiers",
            ["SpecifierSet", "InvalidSpecifier"],
        ),
        FromImport(
            "importlib", "metadata"
        ),
    ],
)


class EnvironmentDetails(Prefab, kw_only=True):
    requires_python: str | None
    dependencies: list[str]
    tool_table: dict[str, str]

    @property
    def requires_python_spec(self):
        return _laz.SpecifierSet(self.requires_python) if self.requires_python else None

    @property
    def dependencies_spec(self):
        return [_laz.Requirement(dep) for dep in self.dependencies]

    def errors(self) -> list[str]:
        error_details = []

        if self.requires_python:
            try:
                _laz.SpecifierSet(self.requires_python)
            except _laz.InvalidSpecifier:
                error_details.append(
                    f"Invalid python version specifier: {self.requires_python!r}"
                )
        for dep in self.dependencies:
            try:
                _laz.Requirement(dep)
            except _laz.InvalidRequirement:
                error_details.append(f"Invalid dependency specification: {dep!r}")

        return error_details


class EnvironmentSpec:
    script_path: str
    raw_spec: str
    lockdata: str | None

    def __init__(
        self,
        script_path: str,
        raw_spec: str,
        *,
        lockdata: str | None = None,
        spec_hash: str | None = None,
        details: EnvironmentDetails | None = None,
    ) -> None:
        self.script_path = script_path
        self.raw_spec = raw_spec
        self.lockdata = lockdata

        self._spec_hash: str | None = spec_hash
        self._lock_hash: str | None = None
        self._details: EnvironmentDetails | None = details

    @classmethod
    def from_script(cls, script_path, lockdata: str | None = None):
        raw_spec = scriptmetadata.parse_file(script_path).blocks.get("script", "")
        return cls(
            script_path=script_path,
            raw_spec=raw_spec,
            lockdata=lockdata
        )

    @property
    def details(self) -> EnvironmentDetails:
        if self._details is None:
            self._details = self.parse_raw()
        return self._details

    @property
    def spec_hash(self) -> str:
        if self._spec_hash is None:
            spec_bytes = self.raw_spec.encode("utf8")
            self._spec_hash = _laz.hashlib.sha3_256(spec_bytes).hexdigest()
        return self._spec_hash
    
    @property
    def lock_hash(self) -> str:
        if self.lockdata and self._lock_hash is None:
            lock_bytes = self.lockdata.encode("utf8")
            self._lock_hash = _laz.hashlib.sha3_256(lock_bytes).hexdigest()
        return self._lock_hash

    def parse_raw(self) -> EnvironmentDetails:
        base_table = _laz.tomllib.loads(self.raw_spec)

        requires_python = base_table.get("requires-python", None)
        dependencies = base_table.get("dependencies", [])

        tool_table = (
            base_table.get("tool", {})
            .get("ducktools", {})
            .get("env", {})
        )

        # noinspection PyArgumentList
        return EnvironmentDetails(
            requires_python=requires_python,
            dependencies=dependencies,
            tool_table=tool_table,
        )

    def generate_lockdata(self, uv_path: str) -> str | None:
        """
        Generate a lockfile from the dependency data
        :param uv_path: Path to the UV executable
        :return: lockfile data as a text string or None if there are no dependencies
        """
        if not self.lockdata:
            # Only go through the process if there is anything to lock
            if deps := "\n".join(self.details.dependencies):
                python_version = []
                if python_spec := self.details.requires_python_spec:
                    # Try to find the minimum python version that satisfies the spec
                    for s in python_spec:
                        if s.operator in {"==", ">=", "~="}:
                            python_version = ["--python-version", s.version]
                            break

                lock_cmd = [
                    uv_path,
                    "pip",
                    "compile",
                    "--universal",
                    "--generate-hashes",
                    "--no-annotate",
                    *python_version,
                    "-",
                ]

                print("Locking dependency tree")
                lock_output = _laz.subprocess.run(
                    lock_cmd,
                    input=deps,
                    capture_output=True,
                    text=True,
                )

                self.lockdata = lock_output.stdout

            else:
                # There are no dependencies - Make a note of this
                # This makes lockdata Truthy
                self.lockdata = "# No Dependencies Declared"

        return self.lockdata

    def as_dict(self):
        return {
            "spec_hash": self.spec_hash,
            "raw_spec": self.raw_spec,
            "lock_hash": self.lock_hash,
            "details": as_dict(self.details),
        }



# ducktools.env
# MIT License
# 
# Copyright (c) 2024 David C Ellis
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import sys
import os
import os.path

from ducktools.lazyimporter import LazyImporter, FromImport, ModuleImport, MultiFromImport
from ducktools.classbuilder.prefab import Prefab, attribute

from . import PROJECT_NAME
from .config import Config, log
from .platform_paths import ManagedPaths
from .catalogue import TempCatalogue
from .environment_specs import EnvironmentSpec
from .exceptions import UVUnavailableError


_laz = LazyImporter(
    [
        # stdlib
        ModuleImport("shutil"),
        ModuleImport("subprocess"),
        # third party
        MultiFromImport(
            "packaging.version",
            ["Version", "InvalidVersion"]
        ),
        # internal
        FromImport(".bundle", "create_bundle"),
        MultiFromImport(
            ".scripts.create_zipapp",
            ["build_env_folder", "build_zipapp"]
        ),
        FromImport(".scripts.get_pip", "retrieve_pip"),
        FromImport(".scripts.get_uv", "retrieve_uv"),
    ],
    globs=globals(),
)


class Manager(Prefab):
    project_name: str = PROJECT_NAME
    config: Config = None

    paths: ManagedPaths = attribute(init=False, repr=False)
    _temp_catalogue: TempCatalogue | None = attribute(default=None, private=True)

    def __prefab_post_init__(self, config):
        self.paths = ManagedPaths(PROJECT_NAME)
        self.config = Config.load(self.paths.config_path) if config is None else config

    @property
    def temp_catalogue(self):
        if self._temp_catalogue is None:
            self._temp_catalogue = TempCatalogue.load(self.paths.cache_db)

            # Clear expired caches on load
            self._temp_catalogue.expire_caches(self.config.cache_lifetime_delta)
        return self._temp_catalogue

    @property
    def is_installed(self):
        return os.path.exists(self.paths.pip_zipapp) and os.path.exists(self.paths.env_folder)

    # Ducktools build commands
    def retrieve_pip(self) -> str:
        return _laz.retrieve_pip(paths=self.paths)

    def retrieve_uv(self, required=False) -> str | None:
        if self.config.use_uv or required:
            uv_path = _laz.retrieve_uv(paths=self.paths)
        else:
            uv_path = None

        if uv_path is None and required:
            raise UVUnavailableError(
                "UV is required for this process but is unavailable"
            )

        return uv_path

    @property
    def install_base_command(self) -> list[str]:
        # Get the installer command for python packages
        # Pip or the faster uv_pip if it is available
        if uv_path := self.retrieve_uv():
            return [uv_path, "pip"]
        else:
            pip_path = self.retrieve_pip()
            return [sys.executable, pip_path, "--disable-pip-version-check"]

    def build_env_folder(self, clear_old_builds=True) -> None:
        _laz.build_env_folder(
            paths=self.paths,
            install_base_command=self.install_base_command,
            clear_old_builds=clear_old_builds,
        )

    def build_zipapp(self, clear_old_builds=True) -> None:
        """Build the ducktools.pyz zipapp"""
        _laz.build_zipapp(
            paths=self.paths,
            install_base_command=self.install_base_command,
            clear_old_builds=clear_old_builds,
        )

    # Install and cleanup commands
    def install(self):
        # Install the ducktools package
        self.build_env_folder(clear_old_builds=True)

    def clear_temporary_cache(self):
        # Clear the temporary environment cache
        log(f"Deleting temporary caches at {self.paths.cache_folder!r}")
        self.temp_catalogue.purge_folder()

    def clear_project_folder(self):
        # Clear the entire ducktools folder
        root_path = self.paths.project_folder
        log(f"Deleting full cache at {root_path!r}")
        _laz.shutil.rmtree(root_path, ignore_errors=True)

    # Script running and bundling commands
    def get_script_env(self, spec: EnvironmentSpec):
        env = self.temp_catalogue.find_or_create_env(
            spec=spec,
            config=self.config,
            uv_path=self.retrieve_uv(),
            installer_command=self.install_base_command,
        )
        return env

    def run_bundled_script(
        self,
        *,
        spec: EnvironmentSpec,
        zipapp_path: str,
        args: list[str],
    ):
        env_vars = {
            "DUCKTOOLS_ENV_LAUNCH_TYPE": "BUNDLE",
            "DUCKTOOLS_ENV_LAUNCH_PATH": zipapp_path,
        }
        self.run_script(
            spec=spec,
            args=args,
            env_vars=env_vars,
        )

    def run_direct_script(
        self,
        *,
        spec: EnvironmentSpec,
        args: list[str],
    ):
        env_vars = {
            "DUCKTOOLS_ENV_LAUNCH_TYPE": "SCRIPT",
            "DUCKTOOLS_ENV_LAUNCH_PATH": spec.script_path,
        }
        self.run_script(
            spec=spec,
            args=args,
            env_vars=env_vars,
        )

    def run_script(
        self,
        *,
        spec: EnvironmentSpec,
        args: list[str],
        env_vars: dict[str, str] | None = None,
    ) -> None:
        """Execute the provided script file with the given arguments

        :param spec: EnvironmentSpec
        :param args: arguments to be provided to the script file
        :param env_vars: Environment variables to set
        """
        env = self.get_script_env(spec)
        env_vars["DUCKTOOLS_ENV_LAUNCH_ENVIRONMENT"] = env.path
        log(f"Using environment at: {env.path}")

        # Update environment variables for access from subprocess
        os.environ.update(env_vars)
        _laz.subprocess.run([env.python_path, spec.script_path, *args])

    def create_bundle(
        self,
        *,
        spec: EnvironmentSpec,
        output_file: str | None = None,
    ) -> None:
        """Create a zipapp bundle for the provided script file

        :param spec: EnvironmentSpec
        :param output_file: output path to zipapp bundle (script_file.pyz default)
        """
        if not self.is_installed:
            self.install()

        _laz.create_bundle(
            script_file=spec.script_path,
            output_file=output_file,
            paths=self.paths,
            installer_command=self.install_base_command,
            lockdata=spec.lockdata,
        )



# ducktools.env
# MIT License
# 
# Copyright (c) 2024 David C Ellis
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import sys

import argparse
from ducktools.lazyimporter import LazyImporter, FromImport

from ducktools.env import __version__, PROJECT_NAME
from ducktools.env.exceptions import UVUnavailableError

_laz = LazyImporter(
    [
        FromImport("ducktools.env.manager", "Manager"),
        FromImport("ducktools.env.environment_specs", "EnvironmentSpec"),
    ]
)


def main():
    parser = argparse.ArgumentParser(
        prog="ducktools-env",
        description="Script runner and bundler for scripts with inline dependencies",
    )

    parser.add_argument("-V", "--version", action="version", version=__version__)

    subparsers = parser.add_subparsers(dest="command", required=True)

    # 'run' command and args
    run_parser = subparsers.add_parser(
        "run",
        help="Launch the provided python script with inline dependencies",
    )

    run_parser.add_argument("script_filename", help="Path to the script to run")

    run_lock_group = run_parser.add_mutually_exclusive_group()
    run_lock_group.add_argument(
        "--with-lock",
        help="Include a lockfile to use when running the script",
        action="store"
    )
    run_lock_group.add_argument(
        "--generate-lock",
        help="Generate a lockfile based on the dependencies in the script",
        action="store_true",
    )

    # 'bundle' command and args
    bundle_parser = subparsers.add_parser(
        "bundle",
        help="Bundle the provided python script with inline dependencies into a python zipapp",
    )

    bundle_parser.add_argument("script_filename", help="Path to the script to bundle into a zipapp")
    bundle_parser.add_argument(
        "-o", "--output",
        help="Output to given filename",
        action="store",
    )

    bundle_lock_group = bundle_parser.add_mutually_exclusive_group()
    bundle_lock_group.add_argument(
        "--with-lock",
        help="Include a lockfile to use when running the script",
        action="store"
    )
    bundle_lock_group.add_argument(
        "--generate-lock",
        help="Generate a lockfile to use when unbundling",
        action="store_true"
    )

    # 'generate_lock' command and args
    generate_lock_parser = subparsers.add_parser(
        "generate_lock",
        help="Generate a lockfile based on inline dependencies in a script"
    )

    generate_lock_parser.add_argument(
        "script_filename",
        help="Path to the script to use to generate a lockfile"
    )

    generate_lock_parser.add_argument(
        "-o", "--output",
        help="Output to given filename",
    )

    # 'clear_cache' command and args
    clear_cache_parser = subparsers.add_parser(
        "clear_cache",
        help="clear the temporary environment cache folder",
    )

    clear_cache_parser.add_argument(
        "--full",
        action="store_true",
        help="clear the full ducktools/env application folder",
    )

    # 'rebuild_env' command and args
    create_zipapp_parser = subparsers.add_parser(
        "rebuild_env",
        help="Recreate the ducktools-env library cache from the installed package"
    )

    create_zipapp_parser.add_argument(
        "--zipapp",
        action="store_true",
        help="Also create the portable ducktools.pyz zipapp",
    )

    args, extras = parser.parse_known_args()

    # Finally create a manager
    manager = _laz.Manager(PROJECT_NAME)

    if args.command == "run":
        spec = _laz.EnvironmentSpec.from_script(
            script_path=args.script_filename
        )
        if args.generate_lock:
            uv_path = manager.retrieve_uv(required=True)
            lockdata = spec.generate_lockdata(uv_path=uv_path)
            lock_path = f"{args.script_filename}.lock"
            with open(lock_path, 'w') as f:
                f.write(lockdata)
        elif lock_path := args.with_lock:
            with open(lock_path, 'r') as f:
                lockdata = f.read()
            spec.lockdata = lockdata

        manager.run_direct_script(
            spec=spec,
            args=extras,
        )
    elif args.command == "bundle":
        if extras:
            arg_text = ' '.join(extras)
            sys.stderr.write(f"Unrecognised arguments: {arg_text}")
            return

        spec = _laz.EnvironmentSpec.from_script(
            script_path=args.script_filename
        )
        
        if args.generate_lock:
            uv_path = manager.retrieve_uv(required=True)
            spec.generate_lockdata(uv_path=uv_path)
        elif lock_path := args.with_lock:
            with open(lock_path, 'r') as f:
                lockdata = f.read()
            spec.lockdata = lockdata

        manager.create_bundle(
            spec=spec,
            output_file=args.output,
        )
    elif args.command == "generate_lock":
        uv_path = manager.retrieve_uv(required=True)
        spec = _laz.EnvironmentSpec.from_script(
            script_path=args.script_filename
        )
        lockdata = spec.generate_lockdata(uv_path=uv_path)

        out_path = args.output if args.output else f"{args.script_filename}.lock"
        with open(out_path, "w") as f:
            f.write(lockdata)

    elif args.command == "clear_cache":
        if extras:
            arg_text = ' '.join(extras)
            sys.stderr.write(f"Unrecognised arguments: {arg_text}")
            return

        if args.full:
            manager.clear_project_folder()
        else:
            manager.clear_temporary_cache()
    elif args.command == "rebuild_env":
        if extras:
            arg_text = ' '.join(extras)
            sys.stderr.write(f"Unrecognised arguments: {arg_text}")
            return

        manager.build_env_folder()
        if args.zipapp:
            manager.build_zipapp()
    else:
        # Should be unreachable
        raise ValueError("Invalid command")


if __name__ == "__main__":
    main()



# DuckTools-EnvMan
# Copyright (C) 2024 David C Ellis
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from ducktools.env import PROJECT_NAME
from ducktools.env.manager import Manager
from ducktools.env.environment_specs import EnvironmentSpec


class TestBuildRetrieve:
    def test_build_retrieve(self, testing_catalogue, test_config):

        manager = Manager(project_name=PROJECT_NAME)

        spec = EnvironmentSpec(
            script_path="path/to/script.py",
            raw_spec="requires-python='>=3.8'\ndependencies=[]\n",
        )

        # Test the env does not exist yet
        assert testing_catalogue.find_env(spec=spec) is None

        real_env = testing_catalogue.find_or_create_env(
            spec=spec,
            config=test_config,
            uv_path=manager.retrieve_uv(),
            installer_command=manager.install_base_command,
        )

        assert real_env is not None

        retrieve_env = testing_catalogue.find_env(spec=spec)

        assert real_env == retrieve_env



# ducktools.env
# MIT License
# 
# Copyright (c) 2024 David C Ellis
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This becomes the bundler bootstrap python script
import sys
import zipfile

from pathlib import Path


# Included in bundle
from _bootstrap import update_libraries, launch_script  # type: ignore
from _vendor import zipp  # type: ignore


def main(script_name):
    # Get updated ducktools and pip
    update_libraries()

    # Get the path to this zipfile and the folder is is being run from
    zip_path = sys.argv[0]

    script_dest = Path(zip_path).with_suffix(".py")

    i = 0
    while script_dest.exists():
        # Keep adding .temp.py until the path doesn't exist - up to a point
        script_dest = script_dest.with_suffix(".temp.py")
        i += 1
        if i > 5:
            raise FileExistsError(
                f"'{script_dest}' already exists, as did all the versions with fewer '.temp' segments"
            )

    working_dir = Path(zip_path).parent

    try:
        with zipfile.ZipFile(zip_path) as zf:
            script_info = zf.getinfo(script_name)
            script_info.filename = script_dest.name

            # Get lockfile if it exists
            lock_name = f"{script_name}.lock"
            try:
                lockdata = zipp.Path(zf, lock_name).read_text()
            except FileNotFoundError:
                # No lockfile
                lockdata = None
            
            # Extract the script file to the existing folder
            zf.extract(script_info, path=working_dir)
        launch_script(
            script_file=str(script_dest),
            zipapp_path=zip_path,
            args=sys.argv[1:],
            lockdata=lockdata,
        )
    finally:
        script_dest.unlink()

# BUNDLE CODE TO EXECUTE SCRIPT FOLLOWS



# ducktools.env
# MIT License
# 
# Copyright (c) 2024 David C Ellis
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import sys
from datetime import datetime, timedelta

import unittest.mock as mock
from pathlib import Path

import pytest

from ducktools.env.catalogue import BaseCatalogue, TempCatalogue, TemporaryEnv
from ducktools.env.catalogue import _laz  # noqa


@pytest.fixture
def mock_save():
    # Mock the .save() function from BaseCatalogue
    with mock.patch.object(BaseCatalogue, "save") as save_func:
        yield save_func


@pytest.fixture
def mock_subprocess():
    # Subprocess module lives in the _laz object
    with mock.patch.object(_laz, "subprocess") as subprocess_obj:
        yield subprocess_obj


@pytest.fixture(scope="function")
def fake_temp_envs(catalogue_path):
    env_0_path = str(Path(catalogue_path).parent / "env_0")
    env_1_path = str(Path(catalogue_path).parent / "env_1")
    python_path = sys.executable

    # ENV examples based on examples folder
    env_0 = TemporaryEnv(
        name="env_0",
        path=env_0_path,
        python_version="3.12.5",
        parent_python=python_path,
        created_on="2024-09-02T14:55:53.102038",
        last_used="2024-09-02T14:55:53.102038",
        spec_hashes=["6986c6ae4a2965a4456333b8c60c5ac923ddca0d7edaa70b36b50f545ed8b24b"],
        installed_modules=[
            "certifi==2024.8.30",
            "charset-normalizer==3.3.2",
            "idna==3.8",
            "markdown-it-py==3.0.0",
            "mdurl==0.1.2",
            "pygments==2.18.0",
            "requests==2.32.3",
            "rich==13.8.0",
            "urllib3==2.2.2",
        ]
    )

    env_1 = TemporaryEnv(
        name="env_1",
        path=env_1_path,
        python_version="3.12.5",
        parent_python=python_path,
        created_on="2024-09-02T14:55:58.827666",
        last_used="2024-09-02T14:55:58.827666",
        spec_hashes=["85cdf5c0f9b109ba70cd936b153fd175307406eb802e05df453d5ccf5a19383f"],
        installed_modules=["cowsay==6.1"],
    )

    return {"env_0": env_0, "env_1": env_1}

@pytest.fixture(scope="function")
def fake_temp_catalogue(catalogue_path, fake_temp_envs):
    cat = TempCatalogue(
        path=catalogue_path,
        environments=fake_temp_envs,
        env_counter=2,
    )

    yield cat


@pytest.mark.usefixtures("mock_save")
class TestTempCatalogue:
    def test_delete_env(self, fake_temp_catalogue, fake_temp_envs, mock_save):
        with mock.patch.object(_laz, "shutil") as shutil_patch:
            rmtree = mock.MagicMock()
            shutil_patch.rmtree = rmtree

            pth = fake_temp_envs["env_0"].path

            fake_temp_catalogue.delete_env("env_0")

            rmtree.assert_called_once_with(pth)

            mock_save.assert_called()

            assert "env_0" not in fake_temp_catalogue.environments

    def test_delete_nonexistent_env(self, fake_temp_catalogue):
        with mock.patch.object(_laz, "shutil"):
            with pytest.raises(FileNotFoundError):
                fake_temp_catalogue.delete_env("env_42")

    def test_purge_folder(self, fake_temp_catalogue, fake_temp_envs):
        with mock.patch.object(_laz, "shutil") as shutil_patch:
            rmtree = mock.MagicMock()
            shutil_patch.rmtree = rmtree

            fake_temp_catalogue.purge_folder()
            rmtree.assert_called_once_with(fake_temp_catalogue.catalogue_folder)

        assert fake_temp_catalogue.environments == {}
        
    def test_oldest_cache(self, fake_temp_catalogue):
        assert fake_temp_catalogue.oldest_cache == "env_0"

        # "Use" env_0
        fake_temp_catalogue.environments["env_0"].last_used = datetime.now().isoformat()
        
        assert fake_temp_catalogue.oldest_cache == "env_1"

        # Empty catalogue returns None as oldest cache
        fake_temp_catalogue.environments = {}

        assert fake_temp_catalogue.oldest_cache is None

    def test_expire_caches(self, fake_temp_catalogue, mock_save):
        with mock.patch.object(fake_temp_catalogue, "delete_env") as del_env:
            # Expire all caches
            fake_temp_catalogue.expire_caches(timedelta(seconds=1))

            calls = [
                mock.call("env_0"),
                mock.call("env_1"),
            ]

            del_env.assert_has_calls(calls)

        mock_save.assert_called_once()


# ducktools.env
# MIT License
#
# Copyright (c) 2024 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import shutil
import subprocess
import sys
import zipapp

from pathlib import Path

import importlib_resources

from . import MINIMUM_PYTHON_STR, bootstrap_requires
from .platform_paths import ManagedPaths
from .exceptions import ScriptNameClash

invalid_script_names = {
    "__main__.py",
    "_bootstrap.py",
    "_platform_paths.py",
    "_vendor.py",
}


def create_bundle(
    *,
    script_file: str,
    output_file: str | None = None,
    paths: ManagedPaths,
    installer_command: list[str],
    lockdata: str | None = None,
) -> None:
    """
    Create a zipapp bundle for the inline script

    :param script_file: path to the source script file
    :param paths: ManagedPaths object containing application path info
    :param installer_command: appropriate UV or PIP 'install' command
    :param output_file: output path for the bundle, if not provided the
                        scriptfile path will be used with `.pyz` added as
                        file extension
    :param lockdata: Content of lockfile or None
    :raises ScriptNameClash: error raised if the script name clashes with a 
                             name required for bootstrapping.
    """
    script_path = Path(script_file)

    if script_path.suffix in {".pyz", ".pyzw"}:
        sys.stderr.write(
            "Bundles must be created from .py scripts not .pyz[w] archives\n"
        )

    if script_path.name in invalid_script_names:
        raise ScriptNameClash(
            f"Script {script_file!r} can't be bundled as the name clashes with "
            f"a script or library required for unbundling"
        )

    with paths.build_folder() as build_folder:
        build_path = Path(build_folder)
        print(f"Building bundle in {build_folder!r}")
        print("Copying libraries into build folder")
        # Don't copy UV - it's platform dependent
        uv_base_exe = "uv.exe" if sys.platform == "win32" else "uv"
        uv_pattern = shutil.ignore_patterns(uv_base_exe, f"{uv_base_exe}.version")

        # Copy pip and ducktools zipapps into folder
        shutil.copytree(
            paths.manager_folder,
            build_path,
            ignore=uv_pattern,
            dirs_exist_ok=True,
        )

        resources = importlib_resources.files("ducktools.env")

        with importlib_resources.as_file(resources) as env_folder:
            platform_paths_path = env_folder / "platform_paths.py"
            bootstrap_path = env_folder / "bootstrapping" / "bootstrap.py"
            main_zipapp_path = env_folder / "bootstrapping" / "bundle_main.py"

            shutil.copy(platform_paths_path, build_path / "_platform_paths.py")
            shutil.copy(bootstrap_path, build_path / "_bootstrap.py")

            # Write __main__.py with script name included
            with open(build_path / "__main__.py", 'w') as main_file:
                main_file.write(main_zipapp_path.read_text())
                main_file.write(f"\nmain({script_path.name!r})\n")

        print("Installing required unpacking libraries")
        vendor_folder = str(build_path / "_vendor")

        # Unpacking libraries use a ducktools determined minimum python
        # This is for the bootstrapping python and not the python that will
        # run the script.
        pip_command = [
            *installer_command,
            "install",
            "-q",
            *bootstrap_requires,
            "--python-version",
            MINIMUM_PYTHON_STR,
            "--only-binary=:all:",
            "--no-compile",
            "--target",
            vendor_folder
        ]

        subprocess.run(pip_command)

        if lockdata:
            # Copy the lockfile to the lock folder
            lock_path = Path(build_path) / f"{script_path.name}.lock"
            lock_path.write_text(lockdata)

        print("Copying script to build folder and bundling")
        shutil.copy(script_path, build_path)

        if output_file is None:
            archive_path = Path(script_file).with_suffix(".pyz")
        else:
            archive_path = Path(output_file)

        zipapp.create_archive(
            source=build_folder,
            target=archive_path,
            interpreter="/usr/bin/env python",
        )

    print(f"Bundled {script_file!r} as '{archive_path}'")
